from ckonlpy.custom_tag import SimpleTemplateTagger
from ckonlpy.dictionary import CustomizedDictionary

class AbstractTagger:

    def __init__(self, dictionary, templates, tagset):
        self.dictionary = dictionary if dictionary else CustomizedDictionary()
        self.template_tagger = SimpleTemplateTagger(self.dictionary, templates, tagset)
        self.tagset = tagset

    def pos(self, phrase, norm=False, stem=False, perfect_match=True):

        def has_None(wordpos_list):
            return len([True for _, pos, _, _ in wordpos_list if pos is None]) > 0

        eojeols = phrase.split()
        tagged = []

        for eojeol in eojeols:

            wordpos_list = self.template_tagger.pos(eojeol, perfect_match)

            if not wordpos_list:
                tagged += self._base.pos(eojeol, norm=norm, stem=stem)
                continue

            for word, pos, _, _ in wordpos_list:
                if pos is not None:
                    tagged.append((word, pos))
                    continue
                for word_, pos_ in self._base.pos(word, norm=norm, stem=stem):
                    tagged.append((word_, pos_))
        return tagged

    def nouns(self, phrase):
        tagged = self.pos(phrase)
        return [w for w, t in tagged if t == 'Noun']

    def morphs(self, phrase, norm=False, stem=False):
        return [s for s, t in self.pos(phrase, norm=norm, stem=stem)]

    def phrases(self, phrase):
        # TODO
        return self._base.phrases(phrase)

    def add_dictionary(self, words, tag, force=False):
        if (not force) and (not (tag in self.tagset)):
            raise ValueError('%s is not available tag' % tag)
        if (force) and (not (tag in self.tagset)):
            self.template_tagger.add_a_template((tag,))
        self.dictionary.add_dictionary(words, tag)

    def load_dictionary(self, fname_list, tag):
        if not (tag in self.tagset):
            raise ValueError('%s is not available tag' % tag)
        self.dictionary.load_dictionary(fname_list, tag)

    def add_a_template(self, a_template):
        self.template_tagger.add_a_template(a_template)

    def set_evaluator(self, my_weight_tuple, my_evaluate_function, test=True):
        self.template_tagger.set_evaluator(my_weight_tuple, my_evaluate_function, test)


from collections import defaultdict

class Postprocessor:
    def __init__(self, base_tagger, stopwords=None,
     passwords=None, passtags=None, replace=None, ngrams=None):
        self.base_tagger = base_tagger
        self.stopwords = stopwords if stopwords else None
        self.passwords = passwords if passwords else None
        self.passtags = passtags if passtags else None
        self.replace = replace if replace else None
        self.ngrams = self._prepare_ngram_converter(ngrams)

    def _prepare_ngram_converter(self, ngrams):
        if not ngrams:
            return None
        first_to_ngram = defaultdict(lambda: [])
        for ngram in ngrams:
            try:
                if isinstance(ngram[0], tuple) and isinstance(ngram[1], str):
                    first_to_ngram[ngram[0][0]].append(ngram)
                elif isinstance(ngram, tuple) and isinstance(ngram[0], str):
                    first_to_ngram[ngram[0]].append((ngram, 'Noun'))
                else:
                    print('ngram format error : {}'.format(ngram))
            except:
                print('ngram format error : {}'.format(ngram))
                continue

        return dict(first_to_ngram)

    def _as_ngram(self, words):
        def same(wordpos_list, ngram_words):
            for wordpos, unigram_word in zip(wordpos_list, ngram_words):
                if (wordpos != unigram_word) and (wordpos[0] != unigram_word):
                    return False
            return True

        words_ = []
        n = len(words)
        idx = 0

        while idx < n:

            candidates = self.ngrams.get(words[idx][0], None)

            if not candidates:
                words_.append(words[idx])
                idx += 1
                continue

            appended = False
            for ngram_words, ngram_pos in candidates:
                stride = len(ngram_words)
                sliced = words[idx:idx+stride]
                if same(sliced, ngram_words):
                    words_.append((' - '.join([w for w, _ in sliced]), ngram_pos))
                    appended = True
                    idx += stride
                    break

            if not appended:
                words_.append(words[idx])
                idx += 1

        return words_

    def pos(self, phrase):
        def to_replace(w):
            if w in self.replace:
                w_ = self.replace[w]
            elif w[0] in self.replace:
                w_ = self.replace[w[0]]
            else:
                return w
            return (w_, w[1]) if isinstance(w_, str) else w_

        words = self.base_tagger.pos(phrase)

        if self.ngrams:
            words = self._as_ngram(words)
        if self.stopwords:
            words = [w for w in words if not ((w in self.stopwords) or (w[0] in self.stopwords))]
        if self.passwords:
            words = [w for w in words if ((w in self.passwords) or (w[0] in self.passwords))]
        if self.passtags:
            words = [w for w in words if w[1] in self.passtags]
        if self.replace:
            words = [to_replace(w) for w in words]
        return words


import glob
import os


installpath = os.path.dirname(os.path.realpath(__file__))

def load_dictionary(directory, encoding='utf-8', ignore_a_syllable=False):
    fnames = glob.glob('%s/*.txt' % directory)
    words = set()
    for fname in fnames:
        try:
            with open(fname, encoding=encoding) as f:
                words_ = {line.strip() for line in f}
                if ignore_a_syllable:
                    words_ = {w for w in words_ if len(w) > 1}
                words.update(words_)
        except Exception as e:
            print(e)
            continue
    return words

def loadtxt(fname, encoding='utf-8'):
    try:
        with open(fname, encoding=encoding) as f:
            return [line.strip() for line in f]
    except Exception as e:
        print(e)
        return []

def load_wordset(fname, encoding='utf-8'):
    try:
        wordset = set()
        with open(fname, encoding=encoding) as f:
            for line in f:
                word = line.strip().split(' ')
                if len(word) == 1:
                    wordset.add(word[0])
                elif len(word) == 2:
                    wordset.add(tuple(word))
        return wordset
    except Exception as e:
        print(e)
        return set()

def load_replace_wordpair(fname, encoding='utf-8'):
    try:
        replace_wordset = {}
        with open(fname, encoding=encoding) as f:
            for line in f:
                try:
                    source, target = line.strip().split('\t')
                    source = source.split(' ')
                    if len(source) == 1:
                        source = source[0]
                    elif len(source) == 2:
                        source = tuple(source)
                    replace_wordset[source] = target
                except:
                    continue
        return replace_wordset
    except Exception as e:
        print(e)
        return {}

def load_ngram(fname, encoding='utf-8'):
    ngrams = []
    with open(fname, encoding=encoding) as f:
        for line in f:
            try:
                if '\t' in line:
                    words, tag = line.strip().split('\t', 1)
                    words = tuple(words.split())
                else:
                    words = tuple(line.split())
                    tag = None
                if len(words) <= 1:
                    continue
                if tag:
                    ngrams.append((words, tag.strip()))
                else:
                    ngrams.append(words)
            except Exception as e:
                print(e)
                continue
    return ngrams


class SimpleEvaluator:
    def __init__(self, preference=None, tagset=None):
        # self.weight = (
        # ('max_length_of_noun', 0.5),
        # ('length_of_phrase', 0.1),
        # ('exist_noun', 0.2),
        # ('single_word', -0.1),
        # ('has_force_tag', 10)
        # )

        ###### customize
        self.weight = (
            ('num_nics', 0.3),
            ('num_nouns', -0.2),
            ('num_names', 0.4),
            ('num_words', -0.5),
            ('num_not_words', -0.7),
            ('len_word', 0.7),
        )

        # preference = { (word, pos) : score }
        self.preference = preference if preference else {}
        self.tagset = tagset if tagset else {}
        self.debug = False

    def select(self, candidates):

        def evaluating_format(wordpos_list):
            formed = (
                wordpos_list,
                wordpos_list[0][2],
                wordpos_list[-1][3],
                self.evaluate(wordpos_list)
            )
            return formed

        def is_overlab(b, e, target):
            return (b < target[2]) and (target[1] < e)

        scoreds = [evaluating_format(c) for c in candidates]

        ## sorting rules. (score, begin index)
        scoreds = sorted(scoreds, key=lambda x: (-x[3], x[1]))

        selected = []
        while scoreds:
            selected.append(scoreds.pop(0))
            b = selected[-1][1]
            e = selected[-1][2]
            scoreds = [c for c in scoreds if not is_overlab(b, e, c)]

        ### sort by begin index
        selected = sorted(selected, key=lambda x: x[1])
        return selected

    ###### customize
    def evaluate(self, candidate):
        num_nics = len([word for word, pos, begin, e in candidate if pos == 'Nic'])  # 닉 단어 개수
        num_nouns = len([word for word, pos, begin, e in candidate if pos == 'Noun'])  # 명사 단어 개수
        num_names = len([word for word, pos, begin, e in candidate if pos == 'Name'])  # 이름 단어 개수
        num_not_words = len(
            [word for word, pos, begin, e in candidate if pos not in ['Noun', 'Nic', 'Name']])  # 명사 닉 이름 외 단어 개수
        len_word = max([e - begin for word, pos, begin, e in candidate])  # 단어중 제일 길이가 긴거

        num_words = len(candidate)  # 모든 단어개수
        has_no_nics = (num_nics != 0)  # 닉여부 확인
        has_no_names = (num_names != 0)  # 이름여부 확인
        has_no_not_words = (num_not_words != 0)  # 닉, 이름, 명사외 품사여부 확인

        scores = (num_nics, num_nouns, num_words, num_names, num_not_words, len_word)

        # 스코어와 웨이트를 곱한것의 합
        score = sum((score * weight for score, (_, weight) in zip(scores, self.weight)))

        if has_no_names:
            for i in [word for word, pos, begin, e in candidate if pos == 'Name']:
                score += 0.5
        if has_no_nics:
            for i in [word for word, pos, begin, e in candidate if pos == 'Nic']:
                score += 0.5
        if has_no_not_words:
            for i in [word for word, pos, begin, e in candidate if pos not in ['Noun', 'Nic', 'Name']]:
                score -= 0.5
                # print('candidate is ', candidate, score)
        return score

        # def evaluate(self, wordpos_list):

        # scores = (
        # _max_length_of_noun(wordpos_list),
        # wordpos_list[-1][3] - wordpos_list[0][2],
        # _num_of_nouns(wordpos_list) > 0,
        # _num_of_words(wordpos_list) == 1,
        # _has_force_tag(wordpos_list, self.tagset)
        # )

        # score_sum = _wordpos_preference(wordpos_list, self.preference)

        # if self.debug:
        # print('\n{}'.format(wordpos_list))
        # for score, (field, weight) in zip(scores, self.weight):
        # print('{}, w={}, s={}, prod={}'.format(
        # field, weight, score, weight * score))
        # print('preference score = {}'.format(score_sum))

        # score_sum += sum((score * weight for score, (_, weight)
        # in zip(scores, self.weight)))

        # return score_sum


def _max_length_of_noun(wordpos_list):
    satisfied = [len(wordpos[0]) for wordpos in wordpos_list if wordpos[1] == 'Noun']
    return max(satisfied) if satisfied else 0


def _num_of_nouns(wordpos_list):
    return len([wordpos for wordpos in wordpos_list if wordpos[1] == 'Noun'])


def _num_of_words(wordpos_list):
    return len(wordpos_list)


def _wordpos_preference(wordpos_list, preference):
    score = 0
    for word, pos, _, _ in wordpos_list:
        score += preference.get((word, pos), 0)
    return score


def _has_force_tag(wordpos_list, tagset):
    return len([wordpos for wordpos in wordpos_list if not (wordpos[1] in tagset)]) > 0


name = "ckonlpy"
__title__ = 'customized-konlpy'
__version__ = '0.0.64'
__author__ = 'Lovit'
__license__ = 'GPL v3'
__copyright__ = 'Copyright 2017 Lovit'

from .dictionary import CustomizedDictionary
from .utils import installpath
from .utils import load_dictionary
from .utils import loadtxt
from .utils import load_wordset
from .utils import load_replace_wordpair
from .utils import load_ngram
from . import custom_tag
from . import tag


twitter = {
    "Adjective": "형용사",
    "Adverb": "부사",
    "Alpha": "알파벳",
    "Conjunction": "접속사",
    "Determiner": "관형사",
    "Eomi": "어미",
    "Exclamation": "감탄사",
    "Foreign": "외국어, 한자 및 기타기호",
    "Hashtag": "트위터 해쉬태그",
    "Josa": "조사",
    "KoreanParticle": "(ex: ㅋㅋ)",
    "Noun": "명사",
    "Number": "숫자",
    "Modifier": "관형사",
    "PreEomi": "선어말어미",
    "Punctuation": "구두점",
    "ScreenName": "트위터 아이디",
    "Suffix": "접미사",
    "Unknown": "미등록어",
    "Verb": "동사",
    "Nic": "닉컴퍼니추가",
    "Name": "사람이름",
}


class CustomizedDictionary:
    def __init__(self, pos2words=None):
        self._pos2words = pos2words if pos2words else {}
        self._max_length = 0
        if pos2words:
            for words in pos2words.values():
                self._max_length = self._get_max_length(
                    words, self._max_length)

    def _get_max_length(self, words, base=0):
        return max(base, max((len(word) for word in words)))

    def add_dictionary(self, words, tag):
        if type(words) == str:
            words = [words]
        wordset = self._pos2words.get(tag, set())
        wordset.update(set(words))
        self._pos2words[tag] = wordset
        self._max_length = self._get_max_length(words, self._max_length)

    def load_dictionary(self, fname_list, tag):
        def load(fname):
            try:
                with open(fname, encoding='utf-8') as f:
                    words = {word.strip() for word in f}
                    return words
            except Exception as e:
                print('load_dictionary error: %s'%e)
                return []

        wordset = self._pos2words.get(tag, set())
        for fname in fname_list:
            wordset.update(load(fname))
        self._pos2words[tag] = wordset

    def get_tags(self, word):
        return tuple(tag for tag, words
            in self._pos2words.items() if word in words)

    def is_tag(self, word, tag):
        return word in self._pos2words.get(tag, {})


from ._abstract import AbstractTagger
from konlpy.tag import Twitter as KoNLPyTwitter
from ckonlpy.custom_tag import SimpleTemplateTagger
from ckonlpy.data.tagset import twitter as tagset
from ckonlpy.dictionary import CustomizedDictionary
from ckonlpy.utils import installpath
from ckonlpy.utils import load_dictionary


class Twitter(AbstractTagger):

    def __init__(self, dictionary=None, templates=None, use_twitter_dictionary=True):
        super().__init__(dictionary, templates, tagset)
        self._base = KoNLPyTwitter()
        self.use_twitter_dictionary = use_twitter_dictionary
        if use_twitter_dictionary:
            self._load_default_dictionary()
        self.template_tagger = SimpleTemplateTagger(self.dictionary, templates)

    def _load_default_dictionary(self):
        directory = '%s/data/twitter/' % installpath
        self.dictionary.add_dictionary(
           load_dictionary('%s/nic' % directory), 'Nic')
        self.dictionary.add_dictionary(
           load_dictionary('%s/name' % directory), 'Name')
        self.dictionary.add_dictionary(
            load_dictionary('%s/noun' % directory, ignore_a_syllable=True), 'Noun')
        # self.dictionary.add_dictionary(
            # load_dictionary('%s/josa' % directory), 'Josa')
        # self.dictionary.add_dictionary(
            # load_dictionary('%s/adverb' % directory), 'Adverb')
        #self.dictionary.add_dictionary(
        #    load_dictionary(modifier_dir), 'Modifier')


from ckonlpy.utils import installpath
from ckonlpy.utils import loadtxt
from ._evaluator import SimpleEvaluator
from pprint import pprint

class SimpleTemplateTagger:

    def __init__(self, dictionary, templates=None, evaluator=None, tagset=None):

        if not evaluator:
            evaluator = SimpleEvaluator(tagset=tagset)

        if not tagset:
            tagset = {}

        self.dictionary = dictionary
        self.templates = _initialize_templates(templates, dictionary, tagset)
        self.evaluator = evaluator
        self.debug = False

    def pos(self, eojeol, perfect_match=True):

        if not eojeol:
            return []

        n = len(eojeol)
        wordpos_nested_list = _match_words(eojeol, self.dictionary)
        matcheds = _match_templates(wordpos_nested_list, self.templates, self.debug)
        if perfect_match:
            matcheds = [matched for matched in matcheds
                if ((matched[0][2] == 0) and (matched[-1][3] == n))]
        if not matcheds:
            return []
        matcheds = self.evaluator.select(matcheds)
        words = _append_unmatched(matcheds, eojeol)
        return words

    def add_a_template(self, a_template):
        if type(a_template) != tuple:
            a_template = tuple(a_template)
        if (a_template in self.templates) == False:
            self.templates.append(a_template)

    def set_evaluator(self, my_weight_tuple, my_evaluate_function, test=True):
        self.evaluator.weight = my_weight_tuple
        self.evaluator.evaluate = my_evaluate_function

        if not test:
            return

        for test_candidate in test_candidates:
            print('candidate')
            pprint(test_candidate)
            print('score = {}\n'.format(self.evaluator.evaluate(test_candidate)))

"""
test_candidates = [
    [('DB손보', 'Nic', 0, 4), ('윤경민', 'Name', 4, 7)],
    [('DB', 'Alpha', 0, 2), ('손보', 'Noun', 2, 4), ('윤경민', 'Name', 4, 7)],
    [('DB', 'Alpha', 0, 2), ('손보', 'Noun', 2, 4), ('윤', 'Noun', 4, 5), ('경민', 'Noun', 5, 7)],
    [('DB손보', 'Nic', 0, 4), ('윤', 'Noun', 4, 5), ('경민', 'Noun', 5, 7),]
]
"""
"""
test_candidates = [
    [('김윤호', 'Name', 0, 4)],
    [('김윤호', 'Norm', 0, 4)],
    [('김', 'Norm', 0, 1),('윤호', 'Name', 1, 4)],
    [('김', 'Norm', 0, 1),('윤호', 'Norm', 1, 4)],
]
"""
test_candidates = [
    [('국민', 'Nic', 0, 2),('최승언', 'Name', 2, 5)],
    [('국민', 'Nic', 0, 2),('최승언', 'Norm', 2, 5)],
    [('국민', 'Norm', 0, 2),('최승언', 'Norm', 2, 5)],
    [('국민', 'Norm', 0, 2),('최승언', 'Name', 2, 5)],
    [('주식', 'Norm', 0, 2),('회사', 'Norm', 2, 4),('닉컴퍼니', 'Norm', 4, 8)],
    [('주식회사', 'Nic', 0, 4),('닉컴퍼니', 'Norm', 4, 8)],
    [('배민', 'Norm', 0, 2),('페이', 'Norm', 2, 4),('머니', 'Norm', 4, 6)],
    [('배민', 'Norm', 0, 2),('페이머니', 'Nic', 2, 6)],
]



def _initialize_templates(templates, dictionary, tagset):
    if not templates:
        templatespath = '%s/data/templates/twitter_templates0' % installpath
        templates = loadtxt(templatespath)
        templates = [tuple(template.split()) for template in templates]

    single_words = [(pos, ) for pos in dictionary._pos2words
                    if not (pos in tagset)]
    templates += [template for template in single_words
                  if not template in templates]

    return templates

def _match_words(eojeol, dictionary):
    n = len(eojeol)

    matched = []
    for b in range(n):
        sublist = []
        for e in range(b+1, min(n, b + dictionary._max_length) + 1):
            word = eojeol[b:e]
            for tag in dictionary.get_tags(word):
                sublist.append((word, tag, b, e))
        matched.append(sublist)
    return matched

def _match_templates(wordpos_nested_list, templates, debug=False):

    n = len(wordpos_nested_list)
    matcheds = []

    # for each begin position
    for wordpos_list in wordpos_nested_list:
        # for each (word, pos, begin, end)
        for wordpos in wordpos_list:
            # for each template
            for template in templates:
                if template[0] == wordpos[1]:
                    # skip a syllable single word
                    if len(template) == 1 and len(wordpos[0]) == 1:
                        continue
                    expandeds = _expand(
                        wordpos, template, wordpos_nested_list, n, debug)
                    matcheds += expandeds

    return matcheds

def _expand(wordpos, template, wordpos_nested_list, n, debug):

    def get_matched_wordpos(wordpos_list, tag):
        return [wordpos for wordpos in wordpos_list if wordpos[1] == tag]

    # Initialize candidates
    candidates = [[wordpos]]

    # Expansion
    for match_tag in template[1:]:
        candidates_ = []
        for candidate in candidates:
            last_index = candidate[-1][3]
            if last_index >= n:
                continue
            expandables = get_matched_wordpos(
                wordpos_nested_list[last_index], match_tag)
            for expandable in expandables:
                expanded = [c for c in candidate] + [expandable]
                candidates_.append(expanded)
        candidates = candidates_

    if debug and candidates:
        print('\ntemplate = {}'.format(template))
        for candidate in candidates:
            print(candidate)

    return candidates

def _append_unmatched(matcheds, eojeol):

    n = len(eojeol)
    words = [word for match, _, _, _ in matcheds for word in match]
    if not words:
        return [(eojeol, None, 0, n)]

    begin = 0
    unmatcheds = []
    for word in words:
        if begin == word[2]:
            begin = word[3]
            continue
        unmatcheds.append((eojeol[begin:word[2]], None, begin, word[2]))
        begin = word[3]
    if begin < n:
        unmatcheds.append((eojeol[begin:], None, begin, n))

    words += unmatcheds
    words = sorted(words, key=lambda x:x[2])
    return words


import ckonlpy
import setuptools
from setuptools import setup, find_packages

with open('README.md', encoding='utf-8') as fh:
    long_description = fh.read()

setuptools.setup(
    name="ats_ckonlpy",
    # version=ckonlpy.__version__,
    # author=ckonlpy.__author__,
    version='1.0.1',
    author='csu5216',
    author_email='csu5216@gmail.com',
    description="KoNLPy wrapping package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/whfh3900/ats_ckonlpy.git',
    packages=setuptools.find_packages(),
    package_data={
        'ckonlpy':['data/*/*.txt', 'data/*/*/*.txt', 'data/templates/*']
    },
    install_requires=["Jpype1>=0.6.1", "konlpy>=0.4.4"],
    classifiers=(
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ),
)
